///// IMPLEMENTATION FILE


#include  "filter.h"
#include    <iostream.h>

// IDAPI error handler ( customize.. )
void DbiShowError(DBIResult=DBIERR_NONE) {}
BOOL Check(DBIResult r)
{
    BOOL b= (r==DBIERR_NONE);
    if (!b) DbiShowError (r);
    return b;
}


#define OFF sizeof(CANExpr)

struct BaseNode::CompileStruct
{
    // For phase 1
    int     NodeCount;
    int     NodeOffset;     // relative to beginning of NodePool
    int     LiteralOffset;  // relative to beginning of LiteralPool

    // For phase 2
    int     FieldNumber;
    int     NodePoolSize;
    char*   Buffer;
    char*   NodePool()      { return Buffer + OFF + NodeOffset; }
    char*   LiteralPool()   { return Buffer + OFF + NodePoolSize + LiteralOffset; }

    BaseNode::CompileStruct(char* b = 0, int nps = 0) : NodeCount(0), NodeOffset(0), LiteralOffset(0), FieldNumber(0), NodePoolSize(nps), Buffer(b) {}
};

// Translation to the IDAPI format (2 passes)
// CompileStruct contains the current offsets :
//  - in the 'node area'.
//  - in the 'literal area'.

// In the first pass, we just have to compute the size of the buffers, thus
// increment the Offsets.

// In the second pass, we also have to fill the buffer.

void UnaryNode::Compile(CompileStruct& data)
{
    SCHECK(Node1);
    CANUnary& CAN   = *(CANUnary*)data.NodePool();
    if (data.Buffer)
    {
        CAN.nodeClass   = nodeUNARY;
        CAN.canOp       = Oper;
    }
    data.NodeOffset += sizeof(CANUnary);
    if (data.Buffer)
    {
        CAN.iOperand1   = data.NodeOffset;
    }
    Node1->Compile(data);
    data.NodeCount++;
}

void BinaryNode::Compile(CompileStruct& data)
{
    SCHECK(Node1);
    SCHECK(Node2);
    CANBinary& CAN  = *(CANBinary*)data.NodePool();
    if (data.Buffer)
    {
        CAN.nodeClass   = nodeBINARY;
        CAN.canOp       = Oper;
    }

    data.NodeOffset += sizeof(CANBinary);

    if (data.Buffer)
    {
        CAN.iOperand1   = data.NodeOffset;
    }
    Node1->Compile(data);

    if (data.Buffer)
    {
        CAN.iOperand2   = data.NodeOffset;
    }
    Node2->Compile(data);

    data.NodeCount++;
}

void CompareNode::Compile(CompileStruct& data)
{
    SCHECK(Node1);
    SCHECK(Node2);
    CANCompare& CAN = *(CANCompare*)data.NodePool();
    if (data.Buffer)
    {
        CAN.nodeClass        = nodeCOMPARE;
        CAN.canOp            = Oper;
        CAN.bCaseInsensitive = CaseInsensitive;
        CAN.iPartialLen      = PartialLen;
    }

    data.NodeOffset += sizeof(CANCompare);

    if (data.Buffer)
    {
        CAN.iOperand1   = data.NodeOffset;
    }
    Node1->Compile(data);

    if (data.Buffer)
    {
        CAN.iOperand2   = data.NodeOffset;
    }
    Node2->Compile(data);

    data.NodeCount++;
}

void FieldNode::Compile(CompileStruct& data)
{
    if (data.Buffer)
    {
        CANField& CAN   = *(CANField*)data.NodePool();
        CAN.nodeClass   = nodeFIELD;
        CAN.canOp       = canFIELD;
        CAN.iFieldNum   =  ++data.FieldNumber;  // ? 
        CAN.iNameOffset = data.LiteralOffset;
        strcpy(data.LiteralPool(), FldName.c_str());
    }
    data.NodeOffset += sizeof(CANField);
    data.LiteralOffset+=FldName.length() + 1;
    data.NodeCount++;
}

void CstNode::Compile(CompileStruct& data)
{
    int s=ValueSize();

    if (data.Buffer)
    {
        CANConst& CAN   = *(CANConst*)data.NodePool();
        CAN.nodeClass   = nodeCONST;
        CAN.canOp       = canCONST;
        CAN.iType       = FldType;
        CAN.iSize       = s;
        CAN.iOffset     = data.LiteralOffset;

        char* buf       = data.LiteralPool();
        memcpy(buf, ValueAddr(), s);
    }

    data.NodeOffset += sizeof(CANConst);

    data.LiteralOffset+=s;
    data.NodeCount++;
}

CANExpr* DBINode::Compile()
{
    if (!_Node) return 0;

    // Compute the size to allocate..
    BaseNode::CompileStruct Data;
    _Node->Compile(Data);

    // Allocate
    int bigsize = OFF + Data.NodeOffset + Data.LiteralOffset;
    CANExpr* CanExpr= (pCANExpr)new char[bigsize];

    // Fill expression specific part
    CANExpr& CAN = *CanExpr;

    CAN.iVer            = CANEXPRVERSION;               
    CAN.iTotalSize      = bigsize;
    CAN.iNodes          = Data.NodeCount;
    CAN.iNodeStart      = OFF;
    CAN.iLiteralStart   = OFF + Data.NodeOffset;

    // Fill node & literal parts
    BaseNode::CompileStruct NewData((char*)CanExpr, Data.NodeOffset);
    _Node->Compile(NewData);

    SCHECK(NewData.NodeOffset == Data.NodeOffset);
    SCHECK(NewData.LiteralPool() == NewData.Buffer + bigsize);

    return CanExpr;
};

// output of a compiled expression
static void DebugNode(ostream& os, const char* nodestart, const char* litstart, const char* dbg)
{
    switch(*(NODEClass*)dbg)
    {
        case nodeUNARY :
        {
            CANUnary& c = *(CANUnary*)dbg;
            switch(c.canOp)
            {
                case canMINUS : os << "-"; break;
                case canNOT   : os << "!"; break;
                case canISBLANK : os << "ISBLANK"; break;
                case canNOTBLANK : os << "NOTBLANK"; break;
                default : SERRORX("Unknown unary op");
            }
            os << "( ";
            DebugNode(os, nodestart, litstart, nodestart + c.iOperand1);
            os << " )";
            break;
        }

        case nodeBINARY:
        {
            CANBinary& c = *(CANBinary*)dbg;
            os << "( ";
            DebugNode(os, nodestart, litstart, nodestart + c.iOperand1);
            os << " ";
            switch(c.canOp)
            {
                case canADD : os << "+"; break;
                case canSUB : os << "-"; break;
                case canMUL : os << "*"; break;
                case canDIV : os << "/"; break;
                case canMOD : os << "%"; break;
                case canAND : os << "AND"; break;
                case canOR  : os << "OR"; break;
                case canEQ  : os << "="; break;
                case canNE  : os << "<>"; break;
                case canGT  : os << ">"; break;
                case canGE  : os << ">="; break;
                case canLT  : os << "<"; break;
                case canLE  : os << "<="; break;
                default     : SERRORX("Unknown binary op");
            }
            os << " ";
            DebugNode(os, nodestart, litstart, nodestart + c.iOperand2);
            os << " )";
            break;
        }

        case nodeCOMPARE:
        {
            CANCompare& c = *(CANCompare*)dbg;
            os << "( ";
            DebugNode(os, nodestart, litstart, nodestart + c.iOperand1);
            os << " ";
            switch(c.canOp)
            {
                case canEQ  : os << "==="; break;
                case canLIKE: os << "LIKE"; break;
                default     : SERRORX("Unknown compare op");
            }
            os << " ";
            DebugNode(os, nodestart, litstart, nodestart + c.iOperand2);
            os << ( c.bCaseInsensitive ? ", insensitive" :", sensitive");
            if ( c.iPartialLen ) os << ", len=" << c.iPartialLen;
            else os << ", len= full";
            os << " )";
            break;
        }

        case nodeFIELD:
        {
            CANField& c = *(CANField*)dbg;
            SCHECK( c.canOp == canFIELD);
            os << "FIELD( ";
            os << "n " << (c.iFieldNum - 1);
            os << " '" << (litstart + c.iNameOffset) << "'";
            os << " )";
            break;
        }

        case nodeCONST:
        {
            CANConst& c = *(CANConst*)dbg;
            SCHECK( c.canOp == canCONST);
            const char* ptr = litstart + c.iOffset;
            os << "( ";
//          os << " size = " << c.iSize << " value : ";

            switch(c.iType)
            {
                case fldINT16 : os << "INT16 = " << *(INT16*)ptr; break;
                case fldINT32 : os << "INT32 = " << *(INT32*)ptr; break;
                case fldFLOAT : os << "FLOAT = " << *(FLOAT*)ptr; break;
                case fldDATE :
                {
                    UINT16 d=0, m=0;
                    INT16 y=0;
                    Check(DbiDateDecode( *(DATE*)ptr, &m, &d, &y));
                    os << "DATE = " << m << '/' << d << '/' << y;
                    break;
                }
                case fldTIME :
                {
                    UINT16 h=0, mn=0, ms=0;
                    Check(DbiTimeDecode( *(TIME*)ptr, &h, &mn, &ms));
                    os << "TIME = " << h << ':' << mn << ':' << ms/1000;
                    break;
                }
                case fldTIMESTAMP :
                {
                    DATE D; TIME T;
                    Check(DbiTimeStampDecode( *(TIMESTAMP*)ptr, &D, &T));
                    UINT16 d=0, m=0;
                    INT16 y=0;
                    Check(DbiDateDecode( D, &m, &d, &y));
                    UINT16 h=0, mn=0, ms=0;
                    Check(DbiTimeDecode( T, &h, &mn, &ms));
                    os << "TIMESTAMP = " << m << '/' << d << '/' << y << ';' << h << ':' << m << ':' << ms/1000;
                    break;
                }
                case fldZSTRING : os << "STRING = " << ptr; break;
                default : SERRORX("Unknown const type");
            }

            os << " )";
            break;
        }

        default: SERRORX("Unknown node type");
    }
}

//#include  <iomanip.h>

ostream& operator << (ostream& os, const CANExpr& CanExpr)
{
    if (!&CanExpr) os << "empty";
    else
    {

    //  os << setbase(10);
    //
    //  os  << "EXPRESSION : "
    //      << " ver=" << CanExpr->iVer
    //      << " totalsize=" << CanExpr->iTotalSize
    //      << " iNodes=" << CanExpr->iNodes
    //      << " nodstart=" << CanExpr->iNodeStart
    //      << " litstart=" << CanExpr->iLiteralStart
    //      << " MAIN = ";

        DebugNode(os,   ((char*)&CanExpr) + CanExpr.iNodeStart,
                        ((char*)&CanExpr) + CanExpr.iLiteralStart,
                        ((char*)&CanExpr) + CanExpr.iNodeStart   );
    }
    return os;
}

// Putting a filter in a cursor should look like this :
#if 0
class _SBASECLASS DBICur
{
    hDBICur     hCur;                // Handle to the cursor
    // ...
    CANExpr*        Filter;
    BOOL            SetFilter(DBINode& dbinode);
};

BOOL DBICur::SetFilter( DBINode& dbinode )
{
    CANExpr* filter = dbinode.Compile();

    // Drop the previous filter ( only one filter at a time in this implementation )
    delete Filter; Filter = filter;
    Check(DbiDropFilter(hCur, 0));

    if (Filter)
    {
        hDBIFilter hf=0;

                                          // put 1 otherwise crash (bug doc)
        if (!Check(DbiAddFilter(hCur, 0L, 1, 0, Filter, 0, &hf)))
        {
            SERRORX(STRING("Bad filter :\n" << *Filter));
            goto err;
        }
        SCHECK(hf);
        if (!Check(DbiActivateFilter(hCur, hf))) goto err;
    }

    return TRUE;

  err:
    delete Filter;
    Filter = 0;
    return FALSE;
}
#endif
